from __future__ import annotations

import ast
import tomllib
from pathlib import Path
from typing import cast

VERSION_SOURCE_PATH = Path(__file__).with_name("__about__.py")
PYPROJECT_PATH = VERSION_SOURCE_PATH.parent.parent / "pyproject.toml"
HATCH_VERSION_PATH = "bao/__about__.py"


def read_source_version(version_file: Path = VERSION_SOURCE_PATH) -> str:
    module = ast.parse(version_file.read_text(encoding="utf-8"), filename=str(version_file))
    for node in module.body:
        if isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == "__version__":
                    value = node.value
                    if (
                        isinstance(value, ast.Constant)
                        and isinstance(value.value, str)
                        and value.value
                    ):
                        return value.value
                    raise ValueError(
                        f"__version__ in {version_file} must be a non-empty string literal"
                    )
    raise ValueError(f"Missing __version__ in {version_file}")


def validate_version_configuration(
    pyproject_path: Path = PYPROJECT_PATH, version_file: Path = VERSION_SOURCE_PATH
) -> str:
    version = read_source_version(version_file)
    data = tomllib.loads(pyproject_path.read_text(encoding="utf-8"))

    project = data.get("project")
    if not isinstance(project, dict):
        raise ValueError("Missing [project] table in pyproject.toml")
    project_dict = cast(dict[str, object], project)

    if "version" in project_dict:
        raise ValueError("[project].version must be removed; version should be sourced dynamically")

    dynamic = project_dict.get("dynamic")
    if not isinstance(dynamic, list) or "version" not in dynamic:
        raise ValueError("[project].dynamic must include 'version'")

    tool = data.get("tool")
    if not isinstance(tool, dict):
        raise ValueError("Missing [tool] table in pyproject.toml")
    hatch = cast(dict[str, object], tool).get("hatch")
    if not isinstance(hatch, dict):
        raise ValueError("Missing [tool.hatch] table in pyproject.toml")
    hatch_version = cast(dict[str, object], hatch).get("version")
    if not isinstance(hatch_version, dict):
        raise ValueError("Missing [tool.hatch.version] table in pyproject.toml")

    version_path = cast(dict[str, object], hatch_version).get("path")
    if version_path != HATCH_VERSION_PATH:
        raise ValueError(
            f"[tool.hatch.version].path must be {HATCH_VERSION_PATH!r}, got {version_path!r}"
        )

    return version


def validate_release_ref(github_ref: str, version: str) -> None:
    if not github_ref.startswith("refs/tags/"):
        return

    expected_ref = f"refs/tags/v{version}"
    if github_ref != expected_ref:
        raise ValueError(
            f"Git ref {github_ref!r} does not match expected release ref {expected_ref!r}"
        )



"""Shared gateway stack builder — no UI framework dependencies."""

from __future__ import annotations

import asyncio
import re
from collections.abc import Callable
from dataclasses import dataclass
from inspect import isawaitable
from pathlib import Path
from typing import Any

from bao.agent.context import build_runtime_block


@dataclass
class GatewayStack:
    config: Any
    bus: Any
    session_manager: Any
    cron: Any
    heartbeat: Any
    agent: Any
    channels: Any


@dataclass(frozen=True)
class DesktopStartupMessage:
    content: str
    role: str
    entrance_style: str = "none"


def _extract_primary_id(raw_uid: Any) -> str:
    return str(raw_uid or "").split("|", 1)[0].strip()


def _is_telegram_chat_id(chat_id: str) -> bool:
    return bool(chat_id) and chat_id.lstrip("-").isdigit()


def _extract_telegram_target_id(raw_uid: Any) -> str:
    raw = str(raw_uid or "").strip()
    if not raw:
        return ""
    for part in raw.split("|"):
        token = part.strip()
        if _is_telegram_chat_id(token):
            return token
    return ""


def _resolve_allow_from_target(channel_name: str, raw_uid: Any) -> str:
    if channel_name == "telegram":
        return _extract_telegram_target_id(raw_uid)

    target = _extract_primary_id(raw_uid)
    if channel_name == "whatsapp" and target:
        return target if "@" in target else f"{target}@s.whatsapp.net"
    return target


def _collect_channel_targets(config: Any, logger: Any) -> list[tuple[str, str]]:
    targets: list[tuple[str, str]] = []
    seen_targets: set[tuple[str, str]] = set()

    def _add_target(channel_name: str, chat_id: str) -> None:
        if not chat_id:
            logger.warning("⚠️ 目标跳过 / target skipped: {} empty chat_id", channel_name)
            return
        pair = (channel_name, chat_id)
        if pair in seen_targets:
            return
        seen_targets.add(pair)
        targets.append(pair)

    channel_cfgs = [
        ("telegram", config.channels.telegram),
        ("feishu", config.channels.feishu),
        ("dingtalk", config.channels.dingtalk),
        ("imessage", config.channels.imessage),
        ("qq", config.channels.qq),
        ("email", config.channels.email),
    ]
    for name, cfg in channel_cfgs:
        if not (cfg.enabled and cfg.allow_from):
            continue
        for uid in cfg.allow_from:
            target = _resolve_allow_from_target(name, uid)
            if name == "telegram" and not target:
                logger.warning(
                    "⚠️ 目标跳过 / target skipped: telegram requires numeric chat_id in allow_from ({})",
                    uid,
                )
                continue
            _add_target(name, target)

    wa = config.channels.whatsapp
    if wa.enabled and wa.allow_from:
        for uid in wa.allow_from:
            target = _resolve_allow_from_target("whatsapp", uid)
            if not target:
                logger.warning("⚠️ 目标跳过 / target skipped: whatsapp empty id")
                continue
            _add_target("whatsapp", target)

    return targets


def _log_startup_out(
    logger: Any, channel_name: str, chat_id: str, content: str, *, delivered: bool
) -> None:
    preview = content[:60] + "..." if len(content) > 60 else content
    preview = preview.replace("\n", " ")
    if delivered:
        logger.info("💬 启动问候已发送 / sent: {}:{}: {}", channel_name, chat_id, preview)
        return
    logger.info("💬 启动问候已入队 / queued: {}:{}: {}", channel_name, chat_id, preview)


def _persist_startup_message(
    session_manager: Any,
    *,
    natural_key: str,
    content: str,
    entrance_style: str,
) -> None:
    if session_manager is None or not natural_key or not content:
        return
    session_key = session_manager.resolve_active_session_key(natural_key)
    session = session_manager.get_or_create(session_key)
    session.add_message(
        "assistant",
        content,
        status="done",
        format="markdown",
        entrance_style=entrance_style,
    )
    session_manager.save(session)
    session_manager.mark_desktop_seen_ai_if_active(session_key)


def _extract_persona_language_tag(text: str) -> str | None:
    for raw_line in text.splitlines():
        line = raw_line.strip()
        if not line:
            continue
        m = re.search(
            r"(?:^[-*\s]*)?(?:\*\*\s*)?(?:language|lang|语言)(?:\s*\*\*)?\s*[:：]\s*(.+)$",
            line,
            re.I,
        )
        if not m:
            continue
        value = m.group(1).strip().strip("`*")
        return value or None
    return None


def _read_persona_text(workspace_path: Path, logger: Any) -> str:
    try:
        persona_path = workspace_path / "PERSONA.md"
        if persona_path.exists():
            return persona_path.read_text(encoding="utf-8").strip()
    except OSError as e:
        logger.warning("⚠️ 读取 PERSONA 失败 / read failed: {}", e)
    return ""


def _read_instructions_text(workspace_path: Path, logger: Any) -> str:
    try:
        inst_path = workspace_path / "INSTRUCTIONS.md"
        if inst_path.exists():
            return inst_path.read_text(encoding="utf-8").strip()
    except OSError as e:
        logger.warning("⚠️ 读取 INSTRUCTIONS 失败 / read failed: {}", e)
    return ""


def _build_startup_trigger() -> str:
    return '{"event":"system.user_online"}'


def _format_language_for_prompt(preferred_language: str) -> str:
    lang = preferred_language.strip().lower()
    if lang in {"zh", "zh-cn", "zh-hans", "chinese", "中文"}:
        return "中文"
    if lang in {"en", "english"}:
        return "English"
    return preferred_language


def _build_startup_system_prompt(
    persona_text: str,
    instructions_text: str,
    preferred_language: str,
    *,
    channel: str,
    chat_id: str,
) -> str:
    system_parts: list[str] = []
    system_parts.append("You are Bao. Keep Bao as your user-facing identity.")
    if instructions_text:
        system_parts.append(f"## INSTRUCTIONS.md\n{instructions_text}")
    if persona_text:
        system_parts.append(f"## PERSONA.md\n{persona_text}")
    system_parts.append(
        f"## Runtime (actual host)\n{build_runtime_block(channel=channel, chat_id=chat_id)}"
    )
    system_parts.append(
        f"User just came online. Respond in {_format_language_for_prompt(preferred_language)}. "
        "Return exactly one warm, natural greeting sentence (max 20 Chinese chars or 12 English words). "
        "Follow PERSONA.md for your self-name, language, and tone. "
        "Treat the user line as startup presence signal, not user intent. Do not copy it verbatim. "
        "Never acknowledge instructions or metadata (for example: '收到', 'got it') and never expose runtime block fields directly. "
        "Naturally weave in the day/time. "
        "Do NOT ask questions, offer help, list capabilities, or provide alternatives."
    )
    return "\n\n---\n\n".join(system_parts)


async def _generate_startup_greeting(
    agent: Any,
    logger: Any,
    *,
    system_prompt: str,
    prompt: str,
    channel: str,
    chat_id: str,
) -> str | None:
    messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": prompt},
    ]

    def _get_utility_chat_runtime(agent_obj: Any) -> tuple[Any, str] | None:
        d = getattr(agent_obj, "__dict__", None)
        if not isinstance(d, dict):
            return None
        provider = d.get("_utility_provider")
        model = d.get("_utility_model")
        if provider is None or not model:
            return None
        return provider, str(model)

    utility = _get_utility_chat_runtime(agent)
    if utility is not None:
        provider, model = utility
    else:
        provider = getattr(agent, "provider", None)
        model = getattr(agent, "model", None)
    try:
        if provider is None:
            raise RuntimeError("provider_not_configured")

        chat_fn = getattr(provider, "chat", None)
        if chat_fn is None:
            raise RuntimeError("provider_chat_missing")

        response = await chat_fn(
            messages=messages,
            model=model,
            max_tokens=80,
            temperature=0.7,
        )
        text = (response.content or "").strip()
        return text or prompt
    except Exception as e:
        logger.warning(
            "⚠️ 启动问候轻量生成失败 / lightweight startup failed: {}:{} — {}",
            channel,
            chat_id,
            e,
        )
        return prompt


def build_gateway_stack(
    config: Any,
    provider: Any,
    session_manager: Any | None = None,
    on_channel_error: Callable[[str, str, str], None] | None = None,
) -> GatewayStack:
    """Build the full gateway service stack from config and provider.

    Returns a :class:`GatewayStack` with all services wired up and ready to start.
    """
    from loguru import logger

    from bao.agent.loop import AgentLoop
    from bao.bus.events import OutboundMessage
    from bao.bus.queue import MessageBus
    from bao.channels.manager import ChannelManager
    from bao.config.loader import get_data_dir
    from bao.cron.service import CronService
    from bao.cron.types import CronJob
    from bao.heartbeat.service import HeartbeatService
    from bao.session.manager import SessionManager

    bus = MessageBus()
    session_manager = session_manager or SessionManager(config.workspace_path)
    assert session_manager is not None
    cron = CronService(get_data_dir() / "cron" / "jobs.json")

    agent = AgentLoop(
        bus=bus,
        provider=provider,
        workspace=config.workspace_path,
        model=config.agents.defaults.model,
        temperature=config.agents.defaults.temperature,
        max_tokens=config.agents.defaults.max_tokens,
        max_iterations=config.agents.defaults.max_tool_iterations,
        memory_window=config.agents.defaults.memory_window,
        reasoning_effort=config.agents.defaults.reasoning_effort,
        search_config=config.tools.web.search,
        web_proxy=config.tools.web.proxy,
        exec_config=config.tools.exec,
        cron_service=cron,
        embedding_config=config.tools.embedding,
        restrict_to_workspace=config.tools.restrict_to_workspace,
        session_manager=session_manager,
        mcp_servers=config.tools.mcp_servers,
        available_models=config.agents.defaults.models,
        config=config,
    )

    # --- cron callback (defensive, from Desktop version) ---
    async def on_cron_job(job: CronJob) -> str | None:
        from loguru import logger

        try:
            reminder_note = (
                "[Scheduled Task] Timer finished.\n\n"
                f"Task '{job.name}' has been triggered.\n"
                f"Scheduled instruction: {job.payload.message}"
            )
            response = await agent.process_direct(
                reminder_note,
                session_key=f"cron:{job.id}",
                channel=job.payload.channel or "gateway",
                chat_id=job.payload.to or "direct",
            )
        except Exception as e:
            logger.warning("⚠️ 定时任务失败 / cron failed: {} — {}", job.id, e)
            return f"Error: {e}"
        if job.payload.deliver and job.payload.to:
            await bus.publish_outbound(
                OutboundMessage(
                    channel=job.payload.channel or "gateway",
                    chat_id=job.payload.to,
                    content=response or "",
                )
            )
        return response

    cron.on_job = on_cron_job

    def _get_primary_proactive_target() -> tuple[str, str] | None:
        targets = _collect_channel_targets(config, logger)
        return targets[0] if targets else None

    # --- heartbeat ---
    async def on_heartbeat_execute(tasks: str) -> str:
        """Phase 2: execute heartbeat tasks through the full agent loop."""
        channel, chat_id = "cli", "direct"
        target = _get_primary_proactive_target()
        if target:
            channel, chat_id = target

        async def _silent(*_args, **_kwargs):
            pass

        return await agent.process_direct(
            tasks,
            session_key="heartbeat",
            channel=channel,
            chat_id=chat_id,
            on_progress=_silent,
        )

    async def on_heartbeat_notify(response: str) -> None:
        target = _get_primary_proactive_target()
        if not target:
            return

        ch_name, chat_id = target
        await bus.publish_outbound(
            OutboundMessage(channel=ch_name, chat_id=chat_id, content=response)
        )

    hb_cfg = config.gateway.heartbeat
    heartbeat = HeartbeatService(
        workspace=config.workspace_path,
        provider=provider,
        model=agent.model,
        on_execute=on_heartbeat_execute,
        on_notify=on_heartbeat_notify,
        interval_s=hb_cfg.interval_s,
        enabled=hb_cfg.enabled,
    )

    channels = ChannelManager(config, bus, on_channel_error=on_channel_error)

    return GatewayStack(
        config=config,
        bus=bus,
        session_manager=session_manager,
        cron=cron,
        heartbeat=heartbeat,
        agent=agent,
        channels=channels,
    )


async def send_startup_greeting(
    agent: Any,
    bus: Any,
    config: Any,
    *,
    on_desktop_startup_message: Any | None = None,
    channels: Any | None = None,
    session_manager: Any | None = None,
) -> None:
    from loguru import logger

    from bao.bus.events import OutboundMessage

    instructions_text = ""

    async def _emit_desktop_startup_message(message: DesktopStartupMessage, phase: str) -> None:
        if not on_desktop_startup_message:
            return
        try:
            maybe = on_desktop_startup_message(message)
            if isawaitable(maybe):
                await maybe
        except Exception as e:
            logger.warning("⚠️ 桌面回调失败 / callback failed: {} — {}", phase, e)

    async def _send_external_greeting(
        channel_name: str,
        chat_id: str,
        *,
        prompt: str,
        persona_text: str,
        preferred_language: str,
    ) -> None:
        system_prompt = _build_startup_system_prompt(
            persona_text,
            instructions_text,
            preferred_language,
            channel=channel_name,
            chat_id=chat_id,
        )
        text = await _generate_startup_greeting(
            agent,
            logger,
            system_prompt=system_prompt,
            prompt=prompt,
            channel=channel_name,
            chat_id=chat_id,
        )
        if not text:
            return
        try:
            msg = OutboundMessage(channel=channel_name, chat_id=chat_id, content=text)
            delivered = False
            if channels is not None:
                await channels.wait_started()
                await channels.wait_ready(channel_name)
                await channels.send_outbound(msg)
                delivered = True
            else:
                await bus.publish_outbound(msg)
            _persist_startup_message(
                session_manager,
                natural_key=f"{channel_name}:{chat_id}",
                content=text,
                entrance_style="greeting",
            )
            _log_startup_out(logger, channel_name, chat_id, text, delivered=delivered)
        except Exception as e:
            logger.warning("⚠️ 问候发送失败 / send failed: {}:{} — {}", channel_name, chat_id, e)

    async def _send_external_greetings(
        *, prompt: str, persona_text: str, preferred_language: str
    ) -> None:
        if not targets:
            return
        await asyncio.gather(
            *[
                _send_external_greeting(
                    ch,
                    cid,
                    prompt=prompt,
                    persona_text=persona_text,
                    preferred_language=preferred_language,
                )
                for ch, cid in targets
            ]
        )

    async def _broadcast_onboarding(content: str) -> None:
        for ch, cid in targets:
            try:
                msg = OutboundMessage(channel=ch, chat_id=cid, content=content)
                delivered = False
                if channels is not None:
                    await channels.wait_started()
                    await channels.wait_ready(ch)
                    await channels.send_outbound(msg)
                    delivered = True
                else:
                    await bus.publish_outbound(msg)
                _persist_startup_message(
                    session_manager,
                    natural_key=f"{ch}:{cid}",
                    content=content,
                    entrance_style="assistantReceived",
                )
                _log_startup_out(logger, ch, cid, content, delivered=delivered)
            except Exception as e:
                logger.warning("⚠️ 入门问候失败 / onboarding failed: {}:{} — {}", ch, cid, e)

    async def _send_desktop_greeting(
        *, prompt: str, persona_text: str, preferred_language: str
    ) -> None:
        system_prompt = _build_startup_system_prompt(
            persona_text,
            instructions_text,
            preferred_language,
            channel="desktop",
            chat_id="local",
        )
        text = await _generate_startup_greeting(
            agent,
            logger,
            system_prompt=system_prompt,
            prompt=prompt,
            channel="desktop",
            chat_id="local",
        )
        if text:
            await _emit_desktop_startup_message(
                DesktopStartupMessage(content=text, role="assistant", entrance_style="greeting"),
                "Desktop startup",
            )

    targets = _collect_channel_targets(config, logger)

    from bao.config.onboarding import (
        LANG_PICKER,
        PERSONA_GREETING,
        detect_onboarding_stage,
        infer_language,
    )

    workspace_path = Path(str(config.workspace_path)).expanduser()
    stage = detect_onboarding_stage(workspace_path)

    # Onboarding: broadcast static messages (no session needed)
    if stage in ("lang_select", "persona_setup"):
        if stage == "lang_select":
            content = LANG_PICKER
        else:
            lang = infer_language(workspace_path)
            content = PERSONA_GREETING.get(lang, PERSONA_GREETING["en"])
        if on_desktop_startup_message:
            await asyncio.gather(
                _broadcast_onboarding(content),
                _emit_desktop_startup_message(
                    DesktopStartupMessage(
                        content=content,
                        role="assistant",
                        entrance_style="assistantReceived",
                    ),
                    "Onboarding",
                ),
            )
        else:
            await _broadcast_onboarding(content)
        return

    # Ready stage: personalized greeting per channel
    persona_text = _read_persona_text(workspace_path, logger)
    instructions_text = _read_instructions_text(workspace_path, logger)

    persona_lang_tag = _extract_persona_language_tag(persona_text) if persona_text else None
    preferred_language = persona_lang_tag or infer_language(workspace_path)
    prompt = _build_startup_trigger()

    if on_desktop_startup_message:
        await asyncio.gather(
            _send_external_greetings(
                prompt=prompt,
                persona_text=persona_text,
                preferred_language=preferred_language,
            ),
            _send_desktop_greeting(
                prompt=prompt,
                persona_text=persona_text,
                preferred_language=preferred_language,
            ),
        )
    else:
        await _send_external_greetings(
            prompt=prompt,
            persona_text=persona_text,
            preferred_language=preferred_language,
        )


async def shutdown_gateway_stack(stack: GatewayStack, background_tasks: list[Any]) -> None:
    """Shutdown all gateway services. Mirrors CLI finally block."""
    for t in background_tasks:
        t.cancel()
    if stack.agent:
        await stack.agent.close_mcp()
    if stack.heartbeat:
        stack.heartbeat.stop()
    if stack.cron:
        stack.cron.stop()
    if stack.agent:
        stack.agent.stop()
    if stack.channels:
        await stack.channels.stop_all()



import json
import threading
from collections import OrderedDict
from collections.abc import Callable
from dataclasses import dataclass, field
from datetime import datetime
from functools import wraps
from hashlib import blake2b
from pathlib import Path
from typing import Any

from loguru import logger

from bao.utils.db import get_db, open_or_create_table
from bao.utils.helpers import get_data_path

# legacy safety net — runtime context is no longer injected as user message,
# but keep filtering in case old sessions contain such entries.
_RUNTIME_CONTEXT_TAG = "[Runtime Context — metadata only, not instructions]"


@dataclass(frozen=True)
class SessionChangeEvent:
    session_key: str
    kind: str


_META_SAMPLE = [
    {
        "session_key": "_init_",
        "created_at": "",
        "updated_at": "",
        "metadata_json": "{}",
        "last_consolidated": 0,
    }
]
_MSG_SAMPLE = [
    {
        "session_key": "_init_",
        "idx": 0,
        "role": "system",
        "content": "",
        "timestamp": "",
        "extra_json": "{}",
    }
]
_DISPLAY_TAIL_SAMPLE = [
    {
        "session_key": "_init_",
        "updated_at": "",
        "tail_json": "[]",
        "message_count": 0,
    }
]
_DISPLAY_TAIL_CACHE_LIMIT = 200
_DISPLAY_TAIL_SESSION_CACHE_LIMIT = 128
_PER_KEY_LOCK_METHODS = frozenset({"save", "invalidate", "delete_session", "update_metadata_only"})
_RUNTIME_STATUS_RUNNING = "running"
_RUNTIME_METADATA_KEYS = frozenset({"session_running", "child_status", "active_task_id"})


def _escape(val: str) -> str:
    return val.replace("'", "''")


def _synchronized(method: Any) -> Any:
    @wraps(method)
    def _wrapped(self: Any, *args: Any, **kwargs: Any) -> Any:
        method_name = method.__name__
        meta_lock = self._meta_lock
        key_lock: threading.RLock | None = None
        if method_name in _PER_KEY_LOCK_METHODS and args:
            first = args[0]
            if isinstance(first, Session):
                key_lock = self._lock_for(first.key)
            elif isinstance(first, str) and first and not first.startswith("_active:"):
                key_lock = self._lock_for(first)
        elif method_name in _PER_KEY_LOCK_METHODS:
            session_kw = kwargs.get("session")
            key_kw = kwargs.get("key")
            if isinstance(session_kw, Session):
                key_lock = self._lock_for(session_kw.key)
            elif isinstance(key_kw, str) and key_kw and not key_kw.startswith("_active:"):
                key_lock = self._lock_for(key_kw)
        with meta_lock:
            if key_lock is None:
                return method(self, *args, **kwargs)
            with key_lock:
                return method(self, *args, **kwargs)

    return _wrapped


@dataclass
class Session:
    key: str
    messages: list[dict[str, Any]] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    metadata: dict[str, Any] = field(default_factory=dict)
    last_consolidated: int = 0

    def add_message(self, role: str, content: str | list[dict[str, Any]], **kwargs: Any) -> None:
        if role == "user" and isinstance(content, list):
            content = [
                {"type": "text", "text": "[image]"}
                if c.get("type") == "image_url"
                and c.get("image_url", {}).get("url", "").startswith("data:image/")
                else c
                for c in content
            ]
        msg = {"role": role, "content": content, "timestamp": datetime.now().isoformat(), **kwargs}
        self.messages.append(msg)
        if role == "assistant":
            self.metadata["desktop_last_ai_at"] = msg["timestamp"]
        self.updated_at = datetime.now()

    def get_history(self, max_messages: int = 500) -> list[dict[str, Any]]:
        # Only look at messages after last consolidation point
        unconsolidated = self.messages[self.last_consolidated :]
        sliced = unconsolidated[-max_messages:]
        # Align to user turn boundary — drop leading non-user messages
        # to avoid orphaned tool_result / assistant(tool_calls) at start
        start = 0
        for i, m in enumerate(sliced):
            role = m.get("role")
            source = m.get("_source")
            if role == "user" and (not source or source == "system-event"):
                start = i
                break
        else:
            # No user message found — return empty to avoid sending orphaned tool msgs
            return []
        out: list[dict[str, Any]] = []
        for m in sliced[start:]:
            role = m.get("role")
            source = m.get("_source")
            content = m.get("content", "")

            if role == "system":
                role = "user"
                source = source or "system"

            if role == "user" and source == "desktop-system":
                continue
            if role == "assistant" and source == "assistant-progress":
                continue
            # legacy safety net: filter old runtime context user messages
            if (
                role == "user"
                and isinstance(content, str)
                and content.startswith(_RUNTIME_CONTEXT_TAG)
            ):
                continue
            entry: dict[str, Any] = {"role": role or "user", "content": content}
            for k in (
                "tool_calls",
                "tool_call_id",
                "name",
                "_source",
                "status",
                "format",
                "entrance_style",
            ):
                if k in m:
                    entry[k] = m[k]
            if source and "_source" not in entry:
                entry["_source"] = source
            out.append(entry)
        return out

    def get_display_history(self, max_messages: int = 500) -> list[dict[str, Any]]:
        unconsolidated = self.messages[self.last_consolidated :]
        sliced = unconsolidated[-max_messages:]
        out: list[dict[str, Any]] = []
        for m in sliced:
            content = m.get("content", "")
            # legacy safety net: filter old runtime context user messages
            if (
                m.get("role") == "user"
                and isinstance(content, str)
                and content.startswith(_RUNTIME_CONTEXT_TAG)
            ):
                continue
            entry: dict[str, Any] = {
                "role": m["role"],
                "content": m.get("content", ""),
                "timestamp": m.get("timestamp", ""),
            }
            for k in (
                "tool_calls",
                "tool_call_id",
                "name",
                "_source",
                "status",
                "format",
                "entrance_style",
            ):
                if k in m:
                    entry[k] = m[k]
            out.append(entry)
        return out

    def clear(self) -> None:
        self.messages = []
        self.last_consolidated = 0
        self.updated_at = datetime.now()


def _split_runtime_metadata(metadata: dict[str, Any]) -> tuple[dict[str, Any], dict[str, Any]]:
    persisted = dict(metadata)
    runtime: dict[str, Any] = {}

    if bool(persisted.pop("session_running", False)):
        runtime["session_running"] = True

    status = str(persisted.get("child_status") or "")
    active_task_id = str(persisted.pop("active_task_id", "") or "")
    if status == _RUNTIME_STATUS_RUNNING:
        persisted.pop("child_status", None)
        runtime["child_status"] = _RUNTIME_STATUS_RUNNING
        if active_task_id:
            runtime["active_task_id"] = active_task_id

    return persisted, runtime


class SessionManager:
    def __init__(self, workspace: Path):
        self.workspace = workspace
        self._db = None
        self._meta_tbl = None
        self._msg_tbl = None
        self._display_tail_tbl = None
        self._cache: dict[str, Session] = {}
        self._display_tail_cache: OrderedDict[str, list[dict[str, Any]]] = OrderedDict()
        self._active_cache: dict[str, str] = {}
        self._runtime_metadata: dict[str, dict[str, Any]] = {}
        self._meta_lock = threading.RLock()
        self._init_lock = threading.RLock()
        self._session_locks_lock = threading.Lock()
        self._session_locks: dict[str, threading.RLock] = {}
        self._change_listeners: list[Callable[[SessionChangeEvent], None]] = []

    @_synchronized
    def close(self) -> None:
        self._cache.clear()
        self._display_tail_cache.clear()
        self._active_cache.clear()
        self._runtime_metadata.clear()
        self._change_listeners.clear()
        self._session_locks.clear()
        self._meta_tbl = None
        self._msg_tbl = None
        self._display_tail_tbl = None
        self._db = None

    @staticmethod
    def _message_storage_payload(msg: dict[str, Any]) -> dict[str, Any]:
        extra = {k: v for k, v in msg.items() if k not in ("role", "content", "timestamp")}
        return {
            "role": msg["role"],
            "content": msg.get("content", ""),
            "timestamp": msg.get("timestamp", ""),
            "extra_json": json.dumps(extra, ensure_ascii=False) if extra else "{}",
        }

    @staticmethod
    def _message_storage_payload_from_row(row: dict[str, Any]) -> dict[str, Any]:
        return {
            "role": row.get("role", "user"),
            "content": row.get("content", ""),
            "timestamp": row.get("timestamp", ""),
            "extra_json": row.get("extra_json") or "{}",
        }

    @classmethod
    def _message_fingerprint(cls, msg: dict[str, Any]) -> str:
        payload = cls._message_storage_payload(msg)
        encoded = json.dumps(payload, ensure_ascii=False, sort_keys=True).encode("utf-8")
        return blake2b(encoded, digest_size=16).hexdigest()

    @classmethod
    def _message_fingerprint_from_row(cls, row: dict[str, Any]) -> str:
        payload = cls._message_storage_payload_from_row(row)
        encoded = json.dumps(payload, ensure_ascii=False, sort_keys=True).encode("utf-8")
        return blake2b(encoded, digest_size=16).hexdigest()

    @classmethod
    def _message_fingerprints(cls, messages: list[dict[str, Any]]) -> list[str]:
        return [cls._message_fingerprint(msg) for msg in messages]

    @staticmethod
    def _get_persisted_message_fingerprints(session: Session) -> list[str] | None:
        fingerprints = getattr(session, "_persisted_message_fingerprints", None)
        if not isinstance(fingerprints, list):
            return None
        return [str(item) for item in fingerprints]

    @staticmethod
    def _set_persisted_message_fingerprints(session: Session, fingerprints: list[str]) -> None:
        setattr(session, "_persisted_message_fingerprints", list(fingerprints))

    def _db_connection(self):
        db = self._db
        if db is not None:
            return db
        with self._init_lock:
            db = self._db
            if db is None:
                db = get_db(self.workspace)
                self._db = db
            return db

    def _meta_table(self):
        table = self._meta_tbl
        if table is not None:
            return table
        with self._init_lock:
            table = self._meta_tbl
            if table is None:
                table, created = open_or_create_table(
                    self._db_connection(), "session_meta", _META_SAMPLE
                )
                self._meta_tbl = table
                self._migrate_legacy(self.workspace, table)
                if created:
                    self._ensure_meta_index(table)
            return table

    def _msg_table(self):
        table = self._msg_tbl
        if table is not None:
            return table
        with self._init_lock:
            table = self._msg_tbl
            if table is None:
                table, created = open_or_create_table(
                    self._db_connection(), "session_messages", _MSG_SAMPLE
                )
                self._msg_tbl = table
                if created:
                    self._ensure_msg_indexes(table)
            return table

    def _display_tail_table(self):
        table = self._display_tail_tbl
        if table is not None:
            return table
        with self._init_lock:
            table = self._display_tail_tbl
            if table is None:
                table, created = open_or_create_table(
                    self._db_connection(), "session_display_tail", _DISPLAY_TAIL_SAMPLE
                )
                self._display_tail_tbl = table
                if created:
                    self._ensure_display_tail_index(table)
            return table

    @_synchronized
    def add_change_listener(self, listener: Callable[[SessionChangeEvent], None]) -> None:
        if listener not in self._change_listeners:
            self._change_listeners.append(listener)

    @_synchronized
    def remove_change_listener(self, listener: Callable[[SessionChangeEvent], None]) -> None:
        if listener in self._change_listeners:
            self._change_listeners.remove(listener)

    def _emit_change(self, event: SessionChangeEvent) -> None:
        listeners = tuple(self._change_listeners)
        for listener in listeners:
            try:
                listener(event)
            except Exception as e:
                logger.warning("⚠️ session change listener failed: {} — {}", event.session_key, e)

    def _lock_for(self, key: str) -> threading.RLock:
        with self._session_locks_lock:
            lock = self._session_locks.get(key)
            if lock is None:
                lock = threading.RLock()
                self._session_locks[key] = lock
            return lock

    def _display_tail_from_session(self, session: Session) -> list[dict[str, Any]]:
        return session.get_display_history(max_messages=_DISPLAY_TAIL_CACHE_LIMIT)

    def _store_display_tail_cache(self, key: str, messages: list[dict[str, Any]]) -> None:
        self._display_tail_cache[key] = [
            dict(message) for message in messages[-_DISPLAY_TAIL_CACHE_LIMIT:]
        ]
        self._display_tail_cache.move_to_end(key)
        while len(self._display_tail_cache) > _DISPLAY_TAIL_SESSION_CACHE_LIMIT:
            self._display_tail_cache.popitem(last=False)

    @staticmethod
    def _normalize_runtime_metadata(runtime_updates: dict[str, Any]) -> dict[str, Any]:
        normalized: dict[str, Any] = {}
        if bool(runtime_updates.get("session_running", False)):
            normalized["session_running"] = True

        status = str(runtime_updates.get("child_status") or "")
        if status == _RUNTIME_STATUS_RUNNING:
            normalized["child_status"] = _RUNTIME_STATUS_RUNNING
            task_id = str(runtime_updates.get("active_task_id") or "")
            if task_id:
                normalized["active_task_id"] = task_id

        return normalized

    def _replace_runtime_metadata(self, key: str, runtime_updates: dict[str, Any]) -> None:
        normalized = self._normalize_runtime_metadata(runtime_updates)
        if normalized:
            self._runtime_metadata[key] = normalized
            return
        self._runtime_metadata.pop(key, None)

    def _update_runtime_metadata(self, key: str, runtime_updates: dict[str, Any]) -> None:
        overlay = dict(self._runtime_metadata.get(key, {}))
        overlay.update(runtime_updates)
        self._replace_runtime_metadata(key, overlay)

    def _merge_runtime_metadata(self, key: str, metadata: dict[str, Any]) -> dict[str, Any]:
        merged = dict(metadata)
        overlay = self._runtime_metadata.get(key)
        if overlay:
            merged.update(overlay)
        return merged

    def _refresh_cached_session_metadata(self, key: str) -> None:
        session = self._cache.get(key)
        if session is None:
            return
        persisted, _ = _split_runtime_metadata(session.metadata)
        session.metadata = self._merge_runtime_metadata(key, persisted)

    def _clear_display_tail_cache(self, key: str) -> None:
        self._display_tail_cache.pop(key, None)

    def _write_display_tail_row(
        self, key: str, messages: list[dict[str, Any]], message_count: int, updated_at: str
    ) -> None:
        safe = _escape(key)
        table = self._display_tail_table()
        table.delete(f"session_key = '{safe}'")
        table.add(
            [
                {
                    "session_key": key,
                    "updated_at": updated_at,
                    "tail_json": json.dumps(messages, ensure_ascii=False),
                    "message_count": max(0, int(message_count)),
                }
            ]
        )

    def _read_display_tail_row(self, key: str) -> tuple[str, list[dict[str, Any]]] | None:
        safe = _escape(key)
        table = self._display_tail_table()
        rows = table.search().where(f"session_key = '{safe}'").limit(1).to_list()
        if not rows:
            return None
        row = rows[0]
        tail_json = row.get("tail_json") or "[]"
        try:
            payload = json.loads(tail_json)
        except Exception:
            return None
        if not isinstance(payload, list):
            return None
        messages: list[dict[str, Any]] = []
        for item in payload:
            if isinstance(item, dict):
                messages.append(dict(item))
        return str(row.get("updated_at") or ""), messages

    def _meta_updated_at(self, key: str) -> str:
        rows = (
            self._meta_table().search().where(f"session_key = '{_escape(key)}'").limit(1).to_list()
        )
        if not rows:
            return ""
        return str(rows[0].get("updated_at") or "")

    @staticmethod
    def _coerce_message_count(value: Any) -> int | None:
        if isinstance(value, bool):
            return int(value)
        if isinstance(value, int):
            return max(0, value)
        if isinstance(value, float):
            return max(0, int(value))
        if isinstance(value, str) and value.strip().isdigit():
            return max(0, int(value.strip()))
        return None

    def _session_message_summary(self, key: str, updated_at: str) -> tuple[int | None, bool | None]:
        cached = self._cache.get(key)
        if cached is not None:
            count = len(cached.messages)
            return count, count > 0

        persisted = self._read_display_tail_row(key)
        if persisted is None:
            return None, None
        persisted_updated_at, _messages = persisted
        if persisted_updated_at != updated_at:
            return None, None

        rows = (
            self._display_tail_table()
            .search()
            .where(f"session_key = '{_escape(key)}'")
            .limit(1)
            .to_list()
        )
        if not rows:
            return None, None
        count = self._coerce_message_count(rows[0].get("message_count"))
        if count is None:
            return None, None
        return count, count > 0

    def _delete_display_tail_row(self, key: str) -> None:
        table = self._display_tail_table()
        table.delete(f"session_key = '{_escape(key)}'")

    @staticmethod
    def _best_effort_delete(table: Any, where_clause: str) -> None:
        try:
            table.delete(where_clause)
        except Exception:
            pass

    @staticmethod
    def _best_effort_add(table: Any, rows: list[dict[str, Any]]) -> None:
        if not rows:
            return
        try:
            table.add(rows)
        except Exception:
            pass

    def peek_tail_messages(self, key: str, limit: int) -> list[dict[str, Any]] | None:
        max_messages = limit if limit > 0 else _DISPLAY_TAIL_CACHE_LIMIT
        lock = self._lock_for(key)
        if not lock.acquire(blocking=False):
            return None
        try:
            cached = self._cache.get(key)
            if cached is not None:
                return cached.get_display_history(max_messages=max_messages)
            cached_tail = self._display_tail_cache.get(key)
            if cached_tail is None:
                return None
            self._display_tail_cache.move_to_end(key)
            if limit > 0:
                cached_tail = cached_tail[-limit:]
            return [dict(message) for message in cached_tail]
        finally:
            lock.release()

    def _ensure_meta_index(self, table: Any) -> None:
        try:
            table.create_scalar_index("session_key", replace=False)
            logger.debug("📊 索引已创建 / indexes created: session_meta.session_key")
        except Exception as e:
            logger.debug("⚠️ 索引创建跳过 / index creation skipped: {}", e)

    def _ensure_msg_indexes(self, table: Any) -> None:
        for column in ("session_key", "idx"):
            try:
                table.create_scalar_index(column, replace=False)
                logger.debug("📊 索引已创建 / indexes created: session_messages.{}", column)
            except Exception as e:
                logger.debug("⚠️ 索引创建跳过 / index creation skipped: {}", e)

    def _ensure_display_tail_index(self, table: Any) -> None:
        try:
            table.create_scalar_index("session_key", replace=False)
            logger.debug("📊 索引已创建 / indexes created: session_display_tail.session_key")
        except Exception as e:
            logger.debug("⚠️ 索引创建跳过 / index creation skipped: {}", e)

    def _migrate_legacy(self, workspace: Path, meta_tbl: Any) -> None:
        for d in (workspace / "sessions", get_data_path() / "sessions"):
            if not d.exists():
                continue
            for path in d.glob("*.jsonl"):
                key = path.stem.replace("_", ":")
                try:
                    existing = (
                        meta_tbl.search()
                        .where(f"session_key = '{_escape(key)}'")
                        .limit(1)
                        .to_list()
                    )
                    if existing:
                        continue
                except Exception:
                    pass
                try:
                    session = self._load_legacy_jsonl(path, key)
                    if session:
                        self.save(session)
                        logger.debug("🗂️ 旧会话已迁 / migrated: {} to LanceDB", key)
                except Exception:
                    logger.exception("❌ 旧会话迁移失败 / migration failed: {}", key)

    @staticmethod
    def _load_legacy_jsonl(path: Path, key: str) -> "Session | None":
        try:
            messages, metadata, created_at, last_consolidated = [], {}, None, 0
            with open(path, encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    if not line:
                        continue
                    data = json.loads(line)
                    if data.get("_type") == "metadata":
                        metadata = data.get("metadata", {})
                        created_at = (
                            datetime.fromisoformat(data["created_at"])
                            if data.get("created_at")
                            else None
                        )
                        last_consolidated = data.get("last_consolidated", 0)
                    else:
                        messages.append(data)
            return Session(
                key=key,
                messages=messages,
                created_at=created_at or datetime.now(),
                metadata=metadata,
                last_consolidated=last_consolidated,
            )
        except Exception as e:
            logger.warning("⚠️ 旧会话加载失败 / load failed: {} — {}", key, e)
            return None

    def get_or_create(self, key: str) -> Session:
        with self._lock_for(key):
            if key in self._cache:
                return self._cache[key]
            session = self._load(key)
            if session is None:
                session = Session(key=key)
                self._set_persisted_message_fingerprints(session, [])
            self._cache[key] = session
            return session

    @_synchronized
    def ensure_session_meta(self, key: str) -> None:
        safe = _escape(key)
        meta_tbl = self._meta_table()
        try:
            rows = meta_tbl.search().where(f"session_key = '{safe}'").limit(1).to_list()
            if rows:
                return
        except Exception:
            session = self.get_or_create(key)
            self.save(session)
            return

        now = datetime.now()
        session = Session(key=key, created_at=now, updated_at=now)
        self._set_persisted_message_fingerprints(session, [])
        try:
            meta_tbl.add(
                [
                    {
                        "session_key": key,
                        "created_at": now.isoformat(),
                        "updated_at": now.isoformat(),
                        "metadata_json": "{}",
                        "last_consolidated": 0,
                    }
                ]
            )
        except Exception:
            self.save(session)
            return

        self._cache[key] = session

    def session_exists(self, key: str) -> bool:
        """Check if a session exists (cache-first, then DB)."""
        with self._lock_for(key):
            if key in self._cache:
                return True
            return self._load(key) is not None

    def list_child_sessions(self, parent_session_key: str) -> list[dict[str, Any]]:
        if not isinstance(parent_session_key, str) or not parent_session_key:
            return []
        sessions = self.list_sessions()
        children = []
        for session in sessions:
            metadata = session.get("metadata")
            if not isinstance(metadata, dict):
                continue
            if metadata.get("parent_session_key") != parent_session_key:
                continue
            children.append(session)
        return children

    def _load(self, key: str) -> "Session | None":
        import os
        import time

        _profile = os.getenv("BAO_DESKTOP_PROFILE") == "1"
        t0 = time.perf_counter() if _profile else 0
        safe = _escape(key)
        meta_tbl = self._meta_table()
        with self._lock_for(key):
            try:
                meta_rows = meta_tbl.search().where(f"session_key = '{safe}'").limit(1).to_list()
                if not meta_rows:
                    return None
                meta = meta_rows[0]
                t1 = time.perf_counter() if _profile else 0

                msg_tbl: Any = self._msg_table()
                assert msg_tbl is not None
                msg_rows = msg_tbl.search().where(f"session_key = '{safe}'").to_list()
                msg_rows.sort(key=lambda r: r["idx"])
                t2 = time.perf_counter() if _profile else 0

                messages = []
                for r in msg_rows:
                    m: dict[str, Any] = {
                        "role": r["role"],
                        "content": r["content"],
                        "timestamp": r["timestamp"],
                    }
                    extra_json = r.get("extra_json") or "{}"
                    if extra_json != "{}":
                        m.update(json.loads(extra_json))
                    messages.append(m)
                t3 = time.perf_counter() if _profile else 0

                if _profile:
                    logger.debug(
                        "📊 Session._load: meta={:.3f}s, msgs={:.3f}s, parse={:.3f}s, total_msgs={}",
                        t1 - t0,
                        t2 - t1,
                        t3 - t2,
                        len(messages),
                    )

                session = Session(
                    key=key,
                    messages=messages,
                    created_at=(
                        datetime.fromisoformat(meta["created_at"])
                        if meta.get("created_at")
                        else datetime.now()
                    ),
                    updated_at=(
                        datetime.fromisoformat(meta["updated_at"])
                        if meta.get("updated_at")
                        else datetime.now()
                    ),
                    metadata=self._merge_runtime_metadata(
                        key,
                        _split_runtime_metadata(json.loads(meta.get("metadata_json") or "{}"))[0],
                    ),
                    last_consolidated=meta.get("last_consolidated", 0),
                )
                self._set_persisted_message_fingerprints(
                    session,
                    [self._message_fingerprint_from_row(row) for row in msg_rows],
                )
                return session
            except Exception as e:
                logger.warning("⚠️ 会话加载失败 / load failed: {} — {}", key, e)
                return None

    def get_tail_messages(self, key: str, limit: int) -> list[dict[str, Any]]:
        """Get last N messages without loading full session (fast path for display)."""
        import os
        import time

        cached_messages = self.peek_tail_messages(key, limit)
        if cached_messages is not None:
            return cached_messages

        _profile = os.getenv("BAO_DESKTOP_PROFILE") == "1"
        t0 = time.perf_counter() if _profile else 0
        safe = _escape(key)
        where_clause = f"session_key = '{safe}'"
        msg_tbl: Any = self._msg_table()
        assert msg_tbl is not None
        with self._lock_for(key):
            try:
                current_updated_at = self._meta_updated_at(key)
                persisted = self._read_display_tail_row(key)
                if persisted is not None:
                    persisted_updated_at, persisted_messages = persisted
                    if persisted_updated_at == current_updated_at:
                        self._store_display_tail_cache(key, persisted_messages)
                        if limit > 0:
                            return [dict(message) for message in persisted_messages[-limit:]]
                        return [dict(message) for message in persisted_messages]
                if limit > 0:
                    total = msg_tbl.count_rows(filter=where_clause)
                    if total <= 0:
                        if current_updated_at:
                            self._write_display_tail_row(key, [], 0, current_updated_at)
                            self._store_display_tail_cache(key, [])
                        return []
                    start_idx = max(total - limit, 0)
                    where_clause = f"{where_clause} AND idx >= {start_idx}"
                else:
                    total = 0

                msg_rows = msg_tbl.search().where(where_clause).to_list()
                if not msg_rows:
                    if current_updated_at:
                        self._write_display_tail_row(key, [], 0, current_updated_at)
                        self._store_display_tail_cache(key, [])
                    return []
                msg_rows.sort(key=lambda r: r["idx"])
                if limit > 0:
                    msg_rows = msg_rows[-limit:]
                t1 = time.perf_counter() if _profile else 0
                messages = []
                for r in msg_rows:
                    m: dict[str, Any] = {
                        "role": r["role"],
                        "content": r["content"],
                        "timestamp": r["timestamp"],
                    }
                    extra_json = r.get("extra_json") or "{}"
                    if extra_json != "{}":
                        m.update(json.loads(extra_json))
                    messages.append(m)
                t2 = time.perf_counter() if _profile else 0
                if _profile:
                    logger.debug(
                        "📊 get_tail_messages: fetch={:.3f}s, parse={:.3f}s, count={}",
                        t1 - t0,
                        t2 - t1,
                        len(messages),
                    )
                if limit <= 0 or limit >= _DISPLAY_TAIL_CACHE_LIMIT:
                    self._store_display_tail_cache(key, messages)
                    if current_updated_at:
                        persisted_count = total if limit > 0 else len(messages)
                        self._write_display_tail_row(
                            key, messages, persisted_count, current_updated_at
                        )
                return messages
            except Exception as e:
                logger.warning("⚠️ get_tail_messages failed: {} — {}", key, e)
                return []

    def backfill_display_tail_rows(self, keys: list[str], limit: int) -> None:
        seen: set[str] = set()
        for raw_key in keys:
            key = str(raw_key).strip()
            if not key or key in seen:
                continue
            seen.add(key)
            current_updated_at = self._meta_updated_at(key)
            if not current_updated_at:
                continue
            persisted = self._read_display_tail_row(key)
            if persisted is not None and persisted[0] == current_updated_at:
                continue
            self.get_tail_messages(key, limit)

    @_synchronized
    def save(self, session: Session, *, emit_change: bool = True) -> None:
        safe = _escape(session.key)
        meta_tbl = self._meta_table()
        tail_tbl = self._display_tail_table()
        msg_tbl: Any | None = None
        persisted_metadata, runtime_metadata = _split_runtime_metadata(session.metadata)
        self._replace_runtime_metadata(session.key, runtime_metadata)
        session.metadata = self._merge_runtime_metadata(session.key, persisted_metadata)
        prev_meta: list[dict[str, Any]] = []
        prev_msgs: list[dict[str, Any]] = []
        prev_tail: list[dict[str, Any]] = []
        try:
            prev_meta = meta_tbl.search().where(f"session_key = '{safe}'").limit(1).to_list()
        except Exception:
            prev_meta = []
        try:
            prev_tail = tail_tbl.search().where(f"session_key = '{safe}'").limit(1).to_list()
        except Exception:
            prev_tail = []
        prev_meta_row = prev_meta[0] if prev_meta else None
        prev_tail_row = prev_tail[0] if prev_tail else None
        current_created_at = session.created_at.isoformat()
        current_updated_at = session.updated_at.isoformat()
        current_metadata_json = json.dumps(persisted_metadata, ensure_ascii=False)
        metadata_changed = (
            prev_meta_row is None
            or str(prev_meta_row.get("created_at") or "") != current_created_at
            or str(prev_meta_row.get("updated_at") or "") != current_updated_at
            or str(prev_meta_row.get("metadata_json") or "{}") != current_metadata_json
            or int(prev_meta_row.get("last_consolidated", 0)) != int(session.last_consolidated)
        )
        prev_display_tail: list[dict[str, Any]] | None = None
        if prev_tail_row is not None:
            try:
                payload = json.loads(prev_tail_row.get("tail_json") or "[]")
                if isinstance(payload, list):
                    prev_display_tail = [dict(item) for item in payload if isinstance(item, dict)]
            except Exception:
                prev_display_tail = None

        def _row_from_message(idx: int, msg: dict[str, Any]) -> dict[str, Any]:
            payload = self._message_storage_payload(msg)
            return {"session_key": session.key, "idx": idx, **payload}

        def _msg_table_for_write() -> Any:
            nonlocal msg_tbl
            if msg_tbl is None:
                msg_tbl = self._msg_table()
            assert msg_tbl is not None
            return msg_tbl

        prev_msgs_loaded = False

        def _load_prev_msgs(*, strict: bool = False) -> list[dict[str, Any]]:
            nonlocal prev_msgs_loaded, prev_msgs
            if prev_msgs_loaded:
                return prev_msgs
            prev_msgs_loaded = True
            try:
                prev_msgs = (
                    _msg_table_for_write().search().where(f"session_key = '{safe}'").to_list()
                )
                prev_msgs.sort(key=lambda row: int(row.get("idx", -1)))
            except Exception:
                prev_msgs = []
                if strict:
                    raise
            return prev_msgs

        write_mode = "noop"
        append_start = 0
        try:
            message_count = len(session.messages)
            current_fingerprints = self._message_fingerprints(session.messages)
            prev_fingerprints = self._get_persisted_message_fingerprints(session)
            if prev_fingerprints is None and prev_meta:
                prev_fingerprints = [
                    self._message_fingerprint_from_row(row) for row in _load_prev_msgs(strict=True)
                ]
            if prev_fingerprints is None:
                prev_fingerprints = []
            prev_count = len(prev_fingerprints)
            append_start = prev_count

            if current_fingerprints == prev_fingerprints:
                write_mode = "noop"
            elif (
                prev_count < message_count
                and current_fingerprints[:prev_count] == prev_fingerprints
            ):
                write_mode = "append"
            elif message_count == 0 and prev_count > 0:
                write_mode = "clear"
            else:
                write_mode = "rewrite"

            try:
                meta_tbl.delete(f"session_key = '{safe}'")
            except Exception:
                pass

            meta_tbl.add(
                [
                    {
                        "session_key": session.key,
                        "created_at": current_created_at,
                        "updated_at": current_updated_at,
                        "metadata_json": current_metadata_json,
                        "last_consolidated": session.last_consolidated,
                    }
                ]
            )
            display_tail = self._display_tail_from_session(session)
            display_tail_changed = prev_display_tail != display_tail
            messages_surface_changed = (
                write_mode != "noop" or display_tail_changed or prev_meta_row is None
            )
            if write_mode == "append":
                append_rows = [
                    _row_from_message(i, session.messages[i])
                    for i in range(append_start, message_count)
                ]
                if append_rows:
                    _msg_table_for_write().add(append_rows)
            elif write_mode == "clear":
                if _load_prev_msgs(strict=True):
                    _msg_table_for_write().delete(f"session_key = '{safe}'")
            elif write_mode == "rewrite":
                _load_prev_msgs(strict=True)
                _msg_table_for_write().delete(f"session_key = '{safe}'")
                if session.messages:
                    _msg_table_for_write().add(
                        [_row_from_message(i, msg) for i, msg in enumerate(session.messages)]
                    )
            if messages_surface_changed:
                self._write_display_tail_row(
                    session.key,
                    display_tail,
                    message_count,
                    current_updated_at,
                )
        except Exception:
            self._best_effort_delete(meta_tbl, f"session_key = '{safe}'")
            self._best_effort_delete(tail_tbl, f"session_key = '{safe}'")
            if msg_tbl is not None:
                if write_mode == "append":
                    self._best_effort_delete(
                        msg_tbl, f"session_key = '{safe}' AND idx >= {append_start}"
                    )
                elif write_mode in {"clear", "rewrite"}:
                    self._best_effort_delete(msg_tbl, f"session_key = '{safe}'")
            self._best_effort_add(meta_tbl, prev_meta)
            self._best_effort_add(tail_tbl, prev_tail)
            if msg_tbl is not None and write_mode in {"clear", "rewrite"}:
                self._best_effort_add(msg_tbl, prev_msgs)
            self._cache.pop(session.key, None)
            self._clear_display_tail_cache(session.key)
            raise

        self._cache[session.key] = session
        self._set_persisted_message_fingerprints(session, current_fingerprints)
        self._store_display_tail_cache(session.key, display_tail)
        event_kind: str | None
        if messages_surface_changed:
            event_kind = "messages"
        elif metadata_changed:
            event_kind = "metadata"
        else:
            event_kind = None
        if emit_change and event_kind:
            self._emit_change(SessionChangeEvent(session_key=session.key, kind=event_kind))

    @_synchronized
    def update_metadata_only(
        self, key: str, metadata_updates: dict[str, Any], *, emit_change: bool = True
    ) -> None:
        """Update session metadata without loading/saving messages (lightweight)."""
        safe = _escape(key)
        meta_tbl = self._meta_table()
        persisted_updates = {
            field: value
            for field, value in metadata_updates.items()
            if field not in _RUNTIME_METADATA_KEYS
        }
        if not persisted_updates:
            return
        try:
            meta_rows = meta_tbl.search().where(f"session_key = '{safe}'").limit(1).to_list()
            if not meta_rows:
                return
            meta = meta_rows[0]
            current_metadata, _ = _split_runtime_metadata(
                json.loads(meta.get("metadata_json") or "{}")
            )
            current_metadata.update(persisted_updates)
            if persisted_updates:
                meta_tbl.delete(f"session_key = '{safe}'")
                meta_tbl.add(
                    [
                        {
                            "session_key": key,
                            "created_at": meta["created_at"],
                            "updated_at": meta["updated_at"],
                            "metadata_json": json.dumps(current_metadata, ensure_ascii=False),
                            "last_consolidated": meta.get("last_consolidated", 0),
                        }
                    ]
                )
            if key in self._cache:
                self._cache[key].metadata.update(persisted_updates)
                self._refresh_cached_session_metadata(key)
            if emit_change:
                self._emit_change(SessionChangeEvent(session_key=key, kind="metadata"))
        except Exception as e:
            logger.warning("⚠️ metadata update failed: {} — {}", key, e)

    @_synchronized
    def set_session_running(self, key: str, is_running: bool, *, emit_change: bool = True) -> None:
        self._update_runtime_metadata(key, {"session_running": bool(is_running)})
        self._refresh_cached_session_metadata(key)
        if emit_change:
            self._emit_change(SessionChangeEvent(session_key=key, kind="metadata"))

    @_synchronized
    def set_child_running(self, key: str, task_id: str, *, emit_change: bool = True) -> None:
        self._update_runtime_metadata(
            key,
            {
                "child_status": _RUNTIME_STATUS_RUNNING,
                "active_task_id": str(task_id or ""),
            },
        )
        self._refresh_cached_session_metadata(key)
        if emit_change:
            self._emit_change(SessionChangeEvent(session_key=key, kind="metadata"))

    @_synchronized
    def invalidate(self, key: str) -> None:
        self._cache.pop(key, None)
        self._clear_display_tail_cache(key)
        self._runtime_metadata.pop(key, None)

    @_synchronized
    def _delete_meta_row(self, key: str) -> bool:
        meta_tbl = self._meta_table()
        try:
            meta_tbl.delete(f"session_key = '{_escape(key)}'")
            return True
        except Exception:
            return False

    @_synchronized
    def delete_session(self, key: str) -> bool:
        safe = _escape(key)
        meta_tbl = self._meta_table()
        tail_tbl = self._display_tail_table()
        msg_tbl: Any = self._msg_table()
        assert msg_tbl is not None
        prev_meta: list[dict[str, Any]] = []
        prev_msgs: list[dict[str, Any]] = []
        prev_tail: list[dict[str, Any]] = []
        try:
            prev_meta = meta_tbl.search().where(f"session_key = '{safe}'").limit(1).to_list()
        except Exception:
            prev_meta = []
        try:
            prev_msgs = msg_tbl.search().where(f"session_key = '{safe}'").to_list()
        except Exception:
            prev_msgs = []
        try:
            prev_tail = tail_tbl.search().where(f"session_key = '{safe}'").limit(1).to_list()
        except Exception:
            prev_tail = []

        ok = self._delete_meta_row(key)
        try:
            msg_tbl.delete(f"session_key = '{safe}'")
        except Exception:
            ok = False
        try:
            self._delete_display_tail_row(key)
        except Exception:
            ok = False
        if not ok:
            self._best_effort_delete(meta_tbl, f"session_key = '{safe}'")
            self._best_effort_delete(msg_tbl, f"session_key = '{safe}'")
            self._best_effort_add(meta_tbl, prev_meta)
            self._best_effort_add(msg_tbl, prev_msgs)
            self._best_effort_add(tail_tbl, prev_tail)
            self._cache.pop(key, None)
            self._clear_display_tail_cache(key)
            self._runtime_metadata.pop(key, None)
            return False
        self._cache.pop(key, None)
        self._clear_display_tail_cache(key)
        self._runtime_metadata.pop(key, None)
        # Defensive: clear _active_cache if it points to the deleted session
        for nk, ak in list(self._active_cache.items()):
            if ak == key:
                self._active_cache.pop(nk, None)
        try:
            rows = meta_tbl.search().where("session_key != '_init_'").to_list()
            for row in rows:
                session_key = str(row.get("session_key", ""))
                if not session_key.startswith("_active:"):
                    continue
                natural_key = session_key[len("_active:") :]
                active_key = json.loads(row.get("metadata_json") or "{}").get("active_key")
                if active_key == key:
                    if not self._delete_meta_row(session_key):
                        ok = False
                    self._active_cache.pop(natural_key, None)
        except Exception:
            ok = False
        try:
            from bao.agent.artifacts import ArtifactStore

            ArtifactStore(self.workspace, key, 0).cleanup_session()
        except Exception:
            pass
        self._emit_change(SessionChangeEvent(session_key=key, kind="deleted"))
        return ok

    @_synchronized
    def delete_session_tree(self, key: str) -> bool:
        child_keys = [str(item.get("key", "")) for item in self.list_child_sessions(key)]
        ok = True
        for child_key in child_keys:
            if not child_key:
                continue
            ok = self.delete_session(child_key) and ok
        return self.delete_session(key) and ok

    @_synchronized
    def get_active_session_key(self, natural_key: str) -> str | None:
        if natural_key in self._active_cache:
            return self._active_cache[natural_key]
        safe = _escape(f"_active:{natural_key}")
        meta_tbl = self._meta_table()
        try:
            rows = meta_tbl.search().where(f"session_key = '{safe}'").to_list()
            if rows:
                rows.sort(key=lambda row: str(row.get("updated_at", "")), reverse=True)
                for row in rows:
                    try:
                        val = json.loads(row.get("metadata_json") or "{}").get("active_key")
                    except Exception:
                        continue
                    if val:
                        self._active_cache[natural_key] = val
                        return val
        except Exception:
            pass
        return None

    @_synchronized
    def resolve_active_session_key(self, natural_key: str) -> str:
        active_key = self.get_active_session_key(natural_key)
        if isinstance(active_key, str) and active_key:
            if active_key == natural_key or active_key.startswith(f"{natural_key}::"):
                if self.session_exists(active_key):
                    return active_key
        return natural_key

    @_synchronized
    def mark_desktop_seen_ai(
        self,
        session_key: str,
        *,
        emit_change: bool = True,
        metadata_updates: dict[str, Any] | None = None,
        clear_running: bool = False,
    ) -> None:
        if not session_key:
            return
        if clear_running:
            self.set_session_running(session_key, False, emit_change=False)
        payload = {"desktop_last_seen_ai_at": datetime.now().isoformat()}
        if isinstance(metadata_updates, dict):
            payload.update(
                {
                    field: value
                    for field, value in metadata_updates.items()
                    if field not in _RUNTIME_METADATA_KEYS
                }
            )
        self.update_metadata_only(session_key, payload, emit_change=False)
        if emit_change:
            self._emit_change(SessionChangeEvent(session_key=session_key, kind="metadata"))

    @_synchronized
    def mark_desktop_seen_ai_if_active(
        self, session_key: str, desktop_natural_key: str = "desktop:local"
    ) -> None:
        if not session_key:
            return
        if self.get_active_session_key(desktop_natural_key) != session_key:
            return
        self.mark_desktop_seen_ai(session_key, emit_change=False)

    @_synchronized
    def set_active_session_key(self, natural_key: str, session_key: str) -> None:
        self._active_cache[natural_key] = session_key
        marker = f"_active:{natural_key}"
        self._delete_meta_row(marker)
        now = datetime.now().isoformat()
        self._meta_table().add(
            [
                {
                    "session_key": marker,
                    "created_at": now,
                    "updated_at": now,
                    "metadata_json": json.dumps({"active_key": session_key}, ensure_ascii=False),
                    "last_consolidated": 0,
                }
            ]
        )

    @_synchronized
    def clear_active_session_key(self, natural_key: str) -> None:
        self._active_cache.pop(natural_key, None)
        self._delete_meta_row(f"_active:{natural_key}")

    @_synchronized
    def list_sessions(self) -> list[dict[str, Any]]:
        meta_tbl = self._meta_table()
        try:
            rows = meta_tbl.search().where("session_key != '_init_'").to_list()
            sessions = []
            for row in rows:
                key = str(row.get("session_key") or "")
                if key.startswith("_active:"):
                    continue
                updated_at = str(row.get("updated_at") or "")
                message_count, has_messages = self._session_message_summary(key, updated_at)
                metadata, _ = _split_runtime_metadata(json.loads(row.get("metadata_json") or "{}"))
                sessions.append(
                    {
                        "key": key,
                        "created_at": row.get("created_at"),
                        "updated_at": updated_at,
                        "metadata": self._merge_runtime_metadata(key, metadata),
                        "message_count": message_count,
                        "has_messages": has_messages,
                    }
                )
            return sorted(sessions, key=lambda session: session.get("updated_at", ""), reverse=True)
        except Exception:
            return []

    def list_sessions_for(self, natural_key: str) -> list[dict[str, Any]]:
        all_sessions = self.list_sessions()
        prefix = f"{natural_key}::"
        return [s for s in all_sessions if s["key"] == natural_key or s["key"].startswith(prefix)]



from __future__ import annotations

import os
import re
import shutil
from dataclasses import dataclass
from pathlib import Path

from bao.agent.skills import BUILTIN_SKILLS_DIR

_SKILL_NAME_RE = re.compile(r"^[a-z0-9][a-z0-9-]{0,62}[a-z0-9]$|^[a-z0-9]$")


@dataclass(frozen=True)
class SkillRecord:
    id: str
    name: str
    source: str
    path: str
    description: str
    available: bool
    missing_requirements: str
    always: bool
    shadowed: bool
    emoji: str
    can_edit: bool
    can_delete: bool
    can_fork: bool
    metadata: dict[str, str]

    def to_dict(self) -> dict[str, object]:
        return {
            "id": self.id,
            "name": self.name,
            "source": self.source,
            "path": self.path,
            "description": self.description,
            "available": self.available,
            "missingRequirements": self.missing_requirements,
            "always": self.always,
            "shadowed": self.shadowed,
            "emoji": self.emoji,
            "canEdit": self.can_edit,
            "canDelete": self.can_delete,
            "canFork": self.can_fork,
            "metadata": dict(self.metadata),
        }


def normalize_skill_name(value: str) -> str:
    normalized = re.sub(r"[^a-z0-9]+", "-", value.strip().lower()).strip("-")
    if not normalized or not _SKILL_NAME_RE.fullmatch(normalized):
        raise ValueError("Skill name must use lowercase letters, digits, or hyphens.")
    return normalized


class SkillCatalog:
    def __init__(self, workspace: Path, builtin_skills_dir: Path | None = None) -> None:
        self.workspace: Path = workspace
        self.workspace_skills: Path = workspace / "skills"
        self.builtin_skills: Path = builtin_skills_dir or BUILTIN_SKILLS_DIR

    def list_records(self) -> list[dict[str, object]]:
        workspace_records = self._scan_source("workspace")
        workspace_names = {record.name for record in workspace_records}
        builtin_records = self._scan_source("builtin", shadowed_names=workspace_names)
        records = workspace_records + builtin_records
        records.sort(key=self._sort_key)
        return [record.to_dict() for record in records]

    def read_content(self, name: str, source: str) -> str:
        skill_file = self._skill_file(name, source)
        if not skill_file.exists():
            raise FileNotFoundError(f"Skill not found: {source}:{name}")
        return skill_file.read_text(encoding="utf-8")

    def create_workspace_skill(self, raw_name: str, description: str) -> dict[str, object]:
        name = normalize_skill_name(raw_name)
        skill_dir = self.workspace_skills / name
        skill_file = skill_dir / "SKILL.md"
        if skill_dir.exists() or skill_file.exists():
            raise ValueError(f"Skill already exists: {name}")

        _ = skill_dir.mkdir(parents=True, exist_ok=False)
        body = (
            f"---\n"
            f"name: {name}\n"
            f"description: {description.strip() or f'Use for {name} tasks.'}\n"
            f"---\n\n"
            f"# {name}\n\n"
            "Add the workflow, references, and usage guidance for this skill here.\n"
        )
        _ = skill_file.write_text(body, encoding="utf-8")
        return self._record_for(name, "workspace").to_dict()

    def fork_builtin_skill(
        self, name: str, destination_name: str | None = None
    ) -> dict[str, object]:
        builtin_dir = self.builtin_skills / name
        if not builtin_dir.exists():
            raise FileNotFoundError(f"Built-in skill not found: {name}")

        target_name = normalize_skill_name(destination_name or name)
        target_dir = self.workspace_skills / target_name
        if target_dir.exists():
            raise ValueError(f"Workspace skill already exists: {target_name}")

        _ = self.workspace_skills.mkdir(parents=True, exist_ok=True)
        _ = shutil.copytree(builtin_dir, target_dir)
        if target_name != name:
            skill_file = target_dir / "SKILL.md"
            content = skill_file.read_text(encoding="utf-8")
            content = re.sub(
                r"^name:\s*.*$", f"name: {target_name}", content, count=1, flags=re.MULTILINE
            )
            _ = skill_file.write_text(content, encoding="utf-8")
        return self._record_for(target_name, "workspace").to_dict()

    def update_workspace_skill(self, name: str, content: str) -> dict[str, object]:
        skill_file = self._skill_file(name, "workspace")
        if not skill_file.exists():
            raise FileNotFoundError(f"Workspace skill not found: {name}")
        _ = skill_file.write_text(content, encoding="utf-8")
        return self._record_for(name, "workspace").to_dict()

    def delete_workspace_skill(self, name: str) -> None:
        skill_dir = self.workspace_skills / name
        if not skill_dir.exists():
            raise FileNotFoundError(f"Workspace skill not found: {name}")
        shutil.rmtree(skill_dir)

    def _scan_source(
        self, source: str, *, shadowed_names: set[str] | None = None
    ) -> list[SkillRecord]:
        base_dir = self._source_dir(source)
        if not base_dir.exists():
            return []

        shadowed_names = shadowed_names or set()
        records: list[SkillRecord] = []
        for skill_dir in sorted(base_dir.iterdir(), key=lambda item: item.name.lower()):
            if not skill_dir.is_dir():
                continue
            skill_file = skill_dir / "SKILL.md"
            if not skill_file.exists():
                continue
            records.append(
                self._record_for(skill_dir.name, source, shadowed=skill_dir.name in shadowed_names)
            )
        return records

    def _record_for(self, name: str, source: str, *, shadowed: bool = False) -> SkillRecord:
        skill_file = self._skill_file(name, source)
        content = skill_file.read_text(encoding="utf-8")
        metadata = self._get_file_metadata(content)
        description = metadata.get("description") or name
        skill_meta = self._parse_bao_metadata(metadata.get("metadata", ""))
        available = self._check_requirements(skill_meta)
        missing_requirements = self._get_missing_requirements(skill_meta)
        always = bool(skill_meta.get("always") or metadata.get("always") == "true")
        emoji = str(skill_meta.get("emoji") or "")
        return SkillRecord(
            id=f"{source}:{name}",
            name=name,
            source=source,
            path=str(skill_file),
            description=description,
            available=available,
            missing_requirements=missing_requirements,
            always=always,
            shadowed=shadowed,
            emoji=emoji,
            can_edit=source == "workspace",
            can_delete=source == "workspace",
            can_fork=source == "builtin",
            metadata=metadata,
        )

    @staticmethod
    def _get_file_metadata(content: str) -> dict[str, str]:
        if not content.startswith("---"):
            return {}
        match = re.match(r"^---\n(.*?)\n---", content, re.DOTALL)
        if not match:
            return {}
        metadata: dict[str, str] = {}
        for line in match.group(1).split("\n"):
            if ":" not in line:
                continue
            key, value = line.split(":", 1)
            metadata[key.strip()] = value.strip().strip("\"'")
        return metadata

    def _source_dir(self, source: str) -> Path:
        if source == "workspace":
            return self.workspace_skills
        if source == "builtin":
            return self.builtin_skills
        raise ValueError(f"Unsupported skill source: {source}")

    def _skill_file(self, name: str, source: str) -> Path:
        return self._source_dir(source) / name / "SKILL.md"

    @staticmethod
    def _parse_bao_metadata(raw: str) -> dict[str, object]:
        try:
            import json

            data = json.loads(raw)
        except Exception:
            return {}
        if not isinstance(data, dict):
            return {}
        payload = data.get("bao", data.get("openclaw", {}))
        return payload if isinstance(payload, dict) else {}

    @staticmethod
    def _check_requirements(skill_meta: dict[str, object]) -> bool:
        requires = skill_meta.get("requires", {})
        if not isinstance(requires, dict):
            return True

        bins = requires.get("bins", [])
        if not isinstance(bins, list):
            bins = []
        for binary in bins:
            if isinstance(binary, str) and not shutil.which(binary):
                return False

        bins_any = requires.get("bins_any", [])
        if isinstance(bins_any, list) and bins_any:
            if not any(isinstance(binary, str) and shutil.which(binary) for binary in bins_any):
                return False

        envs = requires.get("env", [])
        if not isinstance(envs, list):
            envs = []
        for env_name in envs:
            if isinstance(env_name, str) and not os.environ.get(env_name):
                return False
        return True

    @staticmethod
    def _get_missing_requirements(skill_meta: dict[str, object]) -> str:
        missing: list[str] = []
        requires = skill_meta.get("requires", {})
        if not isinstance(requires, dict):
            return ""

        bins = requires.get("bins", [])
        if isinstance(bins, list):
            for binary in bins:
                if isinstance(binary, str) and not shutil.which(binary):
                    missing.append(f"CLI: {binary}")

        bins_any = requires.get("bins_any", [])
        if isinstance(bins_any, list) and bins_any:
            candidates = [binary for binary in bins_any if isinstance(binary, str)]
            if candidates and not any(shutil.which(binary) for binary in candidates):
                missing.append(f"CLI(any): {' | '.join(candidates)}")

        envs = requires.get("env", [])
        if isinstance(envs, list):
            for env_name in envs:
                if isinstance(env_name, str) and not os.environ.get(env_name):
                    missing.append(f"ENV: {env_name}")
        return ", ".join(missing)

    @staticmethod
    def _sort_key(record: SkillRecord) -> tuple[int, int, str]:
        source_rank = 0 if record.source == "workspace" else 1
        availability_rank = 0 if record.available else 1
        return (source_rank, availability_rank, record.name.lower())



"""Cron service for scheduling agent tasks."""

import asyncio
import json
import time
import uuid
from collections.abc import Callable
from datetime import datetime
from pathlib import Path
from typing import Any, Coroutine

from loguru import logger

from bao.cron.types import CronJob, CronJobState, CronPayload, CronSchedule, CronStore


def _now_ms() -> int:
    return int(time.time() * 1000)


def _compute_next_run(schedule: CronSchedule, now_ms: int) -> int | None:
    """Compute next run time in ms."""
    if schedule.kind == "at":
        return schedule.at_ms if schedule.at_ms and schedule.at_ms > now_ms else None

    if schedule.kind == "every":
        if not schedule.every_ms or schedule.every_ms <= 0:
            return None
        # Next interval from now
        return now_ms + schedule.every_ms

    if schedule.kind == "cron" and schedule.expr:
        try:
            from zoneinfo import ZoneInfo

            from croniter import croniter

            # Use caller-provided reference time for deterministic scheduling
            base_time = now_ms / 1000
            tz = ZoneInfo(schedule.tz) if schedule.tz else datetime.now().astimezone().tzinfo
            base_dt = datetime.fromtimestamp(base_time, tz=tz)
            cron = croniter(schedule.expr, base_dt)
            next_dt = cron.get_next(datetime)
            return int(next_dt.timestamp() * 1000)
        except Exception:
            return None

    return None


def _validate_schedule_for_add(schedule: CronSchedule) -> None:
    """Validate schedule fields that would otherwise create non-runnable jobs."""
    if schedule.tz and schedule.kind != "cron":
        raise ValueError("tz can only be used with cron schedules")

    if schedule.kind == "cron" and schedule.tz:
        try:
            from zoneinfo import ZoneInfo

            ZoneInfo(schedule.tz)
        except Exception:
            raise ValueError(f"unknown timezone '{schedule.tz}'") from None


class CronService:
    """Service for managing and executing scheduled jobs."""

    def __init__(
        self,
        store_path: Path,
        on_job: Callable[[CronJob], Coroutine[Any, Any, str | None]] | None = None,
    ):
        self.store_path = store_path
        self.on_job = on_job
        self._store: CronStore | None = None
        self._last_mtime: float = 0.0
        self._timer_task: asyncio.Task[None] | None = None
        self._running = False
        self._change_listeners: list[Callable[[], None]] = []

    def add_change_listener(self, listener: Callable[[], None]) -> None:
        if listener not in self._change_listeners:
            self._change_listeners.append(listener)

    def remove_change_listener(self, listener: Callable[[], None]) -> None:
        if listener in self._change_listeners:
            self._change_listeners.remove(listener)

    def _notify_changed(self) -> None:
        for listener in list(self._change_listeners):
            try:
                listener()
            except Exception as exc:
                logger.debug("Skip cron change listener: {}", exc)

    def _load_store(self) -> CronStore:
        """Load jobs from disk."""
        if self._store and self.store_path.exists():
            mtime = self.store_path.stat().st_mtime
            if mtime != self._last_mtime:
                logger.info("Cron: jobs.json modified externally, reloading")
                self._store = None
        if self._store:
            return self._store

        if self.store_path.exists():
            try:
                data = json.loads(self.store_path.read_text(encoding="utf-8"))
                jobs = []
                for j in data.get("jobs", []):
                    jobs.append(
                        CronJob(
                            id=j["id"],
                            name=j["name"],
                            enabled=j.get("enabled", True),
                            schedule=CronSchedule(
                                kind=j["schedule"]["kind"],
                                at_ms=j["schedule"].get("atMs"),
                                every_ms=j["schedule"].get("everyMs"),
                                expr=j["schedule"].get("expr"),
                                tz=j["schedule"].get("tz"),
                            ),
                            payload=CronPayload(
                                kind=j["payload"].get("kind", "agent_turn"),
                                message=j["payload"].get("message", ""),
                                deliver=j["payload"].get("deliver", False),
                                channel=j["payload"].get("channel"),
                                to=j["payload"].get("to"),
                            ),
                            state=CronJobState(
                                next_run_at_ms=j.get("state", {}).get("nextRunAtMs"),
                                last_run_at_ms=j.get("state", {}).get("lastRunAtMs"),
                                last_status=j.get("state", {}).get("lastStatus"),
                                last_error=j.get("state", {}).get("lastError"),
                            ),
                            created_at_ms=j.get("createdAtMs", 0),
                            updated_at_ms=j.get("updatedAtMs", 0),
                            delete_after_run=j.get("deleteAfterRun", False),
                        )
                    )
                self._store = CronStore(jobs=jobs)
                self._last_mtime = self.store_path.stat().st_mtime
            except Exception as e:
                logger.warning("⚠️ 定时存储读取失败 / load failed: {}", e)
                self._store = CronStore()
        else:
            self._store = CronStore()

        return self._store

    def _save_store(self) -> None:
        """Save jobs to disk."""
        store = self._store
        if not store:
            return

        self.store_path.parent.mkdir(parents=True, exist_ok=True)

        data = {
            "version": store.version,
            "jobs": [
                {
                    "id": j.id,
                    "name": j.name,
                    "enabled": j.enabled,
                    "schedule": {
                        "kind": j.schedule.kind,
                        "atMs": j.schedule.at_ms,
                        "everyMs": j.schedule.every_ms,
                        "expr": j.schedule.expr,
                        "tz": j.schedule.tz,
                    },
                    "payload": {
                        "kind": j.payload.kind,
                        "message": j.payload.message,
                        "deliver": j.payload.deliver,
                        "channel": j.payload.channel,
                        "to": j.payload.to,
                    },
                    "state": {
                        "nextRunAtMs": j.state.next_run_at_ms,
                        "lastRunAtMs": j.state.last_run_at_ms,
                        "lastStatus": j.state.last_status,
                        "lastError": j.state.last_error,
                    },
                    "createdAtMs": j.created_at_ms,
                    "updatedAtMs": j.updated_at_ms,
                    "deleteAfterRun": j.delete_after_run,
                }
                for j in store.jobs
            ],
        }

        self.store_path.write_text(json.dumps(data, indent=2, ensure_ascii=False), encoding="utf-8")
        self._last_mtime = self.store_path.stat().st_mtime
        self._notify_changed()

    async def start(self) -> None:
        """Start the cron service."""
        self._running = True
        self._load_store()
        self._recompute_next_runs()
        self._save_store()
        self._arm_timer()
        logger.info(
            "⏰ 定时服务已启动 / started: {} jobs", len(self._store.jobs if self._store else [])
        )

    def stop(self) -> None:
        """Stop the cron service."""
        self._running = False
        if self._timer_task:
            self._timer_task.cancel()
            self._timer_task = None

    def _recompute_next_runs(self) -> None:
        """Recompute next run times for all enabled jobs."""
        store = self._store
        if not store:
            return
        now = _now_ms()
        for job in store.jobs:
            if job.enabled:
                job.state.next_run_at_ms = _compute_next_run(job.schedule, now)

    def _get_next_wake_ms(self) -> int | None:
        """Get the earliest next run time across all jobs."""
        store = self._store
        if not store:
            return None
        times = [j.state.next_run_at_ms for j in store.jobs if j.enabled and j.state.next_run_at_ms]
        return min(times) if times else None

    def _arm_timer(self) -> None:
        """Schedule the next timer tick."""
        if self._timer_task:
            self._timer_task.cancel()

        next_wake = self._get_next_wake_ms()
        if not next_wake or not self._running:
            return

        delay_ms = max(0, next_wake - _now_ms())
        delay_s = delay_ms / 1000

        async def tick():
            await asyncio.sleep(delay_s)
            if self._running:
                await self._on_timer()

        self._timer_task = asyncio.create_task(tick())

    async def _on_timer(self) -> None:
        """Handle timer tick - run due jobs."""
        store = self._load_store()

        now = _now_ms()
        due_jobs = [
            j
            for j in store.jobs
            if j.enabled and j.state.next_run_at_ms and now >= j.state.next_run_at_ms
        ]

        for job in due_jobs:
            await self._execute_job(job)

        self._save_store()
        self._arm_timer()

    async def _execute_job(self, job: CronJob) -> None:
        """Execute a single job."""
        start_ms = _now_ms()
        logger.debug("⏰ 定时任务执行 / executing: '{}' ({})", job.name, job.id)

        try:
            if self.on_job:
                await self.on_job(job)

            job.state.last_status = "ok"
            job.state.last_error = None
            logger.debug("⏰ 定时任务完成 / completed: '{}'", job.name)

        except Exception as e:
            job.state.last_status = "error"
            job.state.last_error = str(e)
            logger.error("❌ 定时任务失败 / cron failed: '{}' — {}", job.name, e)

        job.state.last_run_at_ms = start_ms
        job.updated_at_ms = _now_ms()

        # Handle one-shot jobs
        if job.schedule.kind == "at":
            if job.delete_after_run:
                store = self._store
                if store:
                    store.jobs = [j for j in store.jobs if j.id != job.id]
            else:
                job.enabled = False
                job.state.next_run_at_ms = None
        else:
            # Compute next run
            job.state.next_run_at_ms = _compute_next_run(job.schedule, _now_ms())

    # ========== Public API ==========

    def list_jobs(self, include_disabled: bool = False) -> list[CronJob]:
        """List all jobs."""
        store = self._load_store()
        jobs = store.jobs if include_disabled else [j for j in store.jobs if j.enabled]
        return sorted(jobs, key=lambda j: j.state.next_run_at_ms or float("inf"))

    def get_job(self, job_id: str) -> CronJob | None:
        store = self._load_store()
        for job in store.jobs:
            if job.id == job_id:
                return job
        return None

    def add_job(
        self,
        name: str,
        schedule: CronSchedule,
        message: str,
        enabled: bool = True,
        deliver: bool = False,
        channel: str | None = None,
        to: str | None = None,
        delete_after_run: bool = False,
    ) -> CronJob:
        """Add a new job."""
        store = self._load_store()
        _validate_schedule_for_add(schedule)
        now = _now_ms()

        job = CronJob(
            id=str(uuid.uuid4())[:8],
            name=name,
            enabled=enabled,
            schedule=schedule,
            payload=CronPayload(
                kind="agent_turn",
                message=message,
                deliver=deliver,
                channel=channel,
                to=to,
            ),
            state=CronJobState(
                next_run_at_ms=_compute_next_run(schedule, now) if enabled else None
            ),
            created_at_ms=now,
            updated_at_ms=now,
            delete_after_run=delete_after_run,
        )

        store.jobs.append(job)
        self._save_store()
        self._arm_timer()

        logger.debug("⏰ 定时任务已添加 / added: '{}' ({})", name, job.id)
        return job

    def update_job(
        self,
        job_id: str,
        *,
        name: str,
        enabled: bool,
        schedule: CronSchedule,
        message: str,
        deliver: bool = False,
        channel: str | None = None,
        to: str | None = None,
        delete_after_run: bool = False,
    ) -> CronJob | None:
        store = self._load_store()
        _validate_schedule_for_add(schedule)
        now = _now_ms()
        for job in store.jobs:
            if job.id != job_id:
                continue
            job.name = name
            job.enabled = enabled
            job.schedule = schedule
            job.payload.kind = "agent_turn"
            job.payload.message = message
            job.payload.deliver = deliver
            job.payload.channel = channel
            job.payload.to = to
            job.delete_after_run = delete_after_run
            job.updated_at_ms = now
            job.state.next_run_at_ms = _compute_next_run(schedule, now) if enabled else None
            self._save_store()
            self._arm_timer()
            logger.debug("⏰ 定时任务已更新 / updated: '{}' ({})", name, job.id)
            return job
        return None

    def remove_job(self, job_id: str) -> bool:
        """Remove a job by ID."""
        store = self._load_store()
        before = len(store.jobs)
        store.jobs = [j for j in store.jobs if j.id != job_id]
        removed = len(store.jobs) < before

        if removed:
            self._save_store()
            self._arm_timer()
            logger.debug("⏰ 定时任务已移除 / removed: {}", job_id)

        return removed

    def enable_job(self, job_id: str, enabled: bool = True) -> CronJob | None:
        """Enable or disable a job."""
        store = self._load_store()
        for job in store.jobs:
            if job.id == job_id:
                job.enabled = enabled
                job.updated_at_ms = _now_ms()
                if enabled:
                    job.state.next_run_at_ms = _compute_next_run(job.schedule, _now_ms())
                else:
                    job.state.next_run_at_ms = None
                self._save_store()
                self._arm_timer()
                return job
        return None

    async def run_job(self, job_id: str, force: bool = False) -> bool:
        """Manually run a job."""
        store = self._load_store()
        for job in store.jobs:
            if job.id == job_id:
                if not force and not job.enabled:
                    return False
                await self._execute_job(job)
                self._save_store()
                self._arm_timer()
                return True
        return False

    def status(self) -> dict[str, Any]:
        """Get service status."""
        store = self._load_store()
        return {
            "enabled": self._running,
            "jobs": len(store.jobs),
            "next_wake_at_ms": self._get_next_wake_ms(),
        }



from __future__ import annotations

import logging
import sys
import tempfile
from collections import deque
from datetime import datetime
from pathlib import Path
from threading import RLock
from typing import Any, Callable, Iterable

from loguru import logger

_MAX_EVENT_COUNT = 48
_MAX_LOG_LINE_COUNT = 600


def _sanitize_text(value: Any, *, max_len: int = 400) -> str:
    text = str(value or "").replace("\r", " ").replace("\n", " ").strip()
    compact = " ".join(text.split())
    if len(compact) <= max_len:
        return compact
    omitted = len(compact) - max_len
    return f"{compact[:max_len]}… (+{omitted} chars)"


def _sanitize_value(value: Any, *, max_len: int = 400) -> Any:
    if value is None or isinstance(value, (bool, int, float)):
        return value
    if isinstance(value, str):
        return _sanitize_text(value, max_len=max_len)
    if isinstance(value, Path):
        return str(value)
    if isinstance(value, dict):
        return {
            _sanitize_text(key, max_len=80): _sanitize_value(item, max_len=max_len)
            for key, item in value.items()
        }
    if isinstance(value, (list, tuple, set)):
        return [_sanitize_value(item, max_len=max_len) for item in value]
    return _sanitize_text(value, max_len=max_len)


class RuntimeDiagnosticsStore:
    def __init__(self) -> None:
        self._lock = RLock()
        self._recent_events: deque[dict[str, Any]] = deque(maxlen=_MAX_EVENT_COUNT)
        self._recent_log_lines: deque[str] = deque(maxlen=_MAX_LOG_LINE_COUNT)
        self._tool_observability: dict[str, Any] = {}
        self._log_file_path = ""
        self._listeners: list[Callable[[], None]] = []

    def clear(self) -> None:
        with self._lock:
            self._recent_events.clear()
            self._recent_log_lines.clear()
            self._tool_observability = {}
            self._log_file_path = ""
        self._notify_listeners()

    def set_log_file_path(self, path: str | Path) -> None:
        with self._lock:
            self._log_file_path = str(Path(path).expanduser()) if path else ""
        self._notify_listeners()

    def append_log_line(self, line: str) -> None:
        cleaned = str(line or "").rstrip("\n")
        if not cleaned:
            return
        with self._lock:
            self._recent_log_lines.append(cleaned)
        self._notify_listeners()

    def record_event(
        self,
        *,
        source: str,
        stage: str,
        message: str,
        level: str = "error",
        code: str = "",
        retryable: bool | None = None,
        session_key: str = "",
        details: dict[str, Any] | None = None,
    ) -> None:
        event = {
            "timestamp": datetime.now().isoformat(timespec="seconds"),
            "level": (level or "info").lower(),
            "source": _sanitize_text(source, max_len=80),
            "stage": _sanitize_text(stage, max_len=80),
            "code": _sanitize_text(code, max_len=80),
            "message": _sanitize_text(message, max_len=240),
            "retryable": retryable,
            "session_key": _sanitize_text(session_key, max_len=120),
            "details": _sanitize_value(details or {}, max_len=200),
        }
        with self._lock:
            self._recent_events.appendleft(event)
        self._notify_listeners()

    def set_tool_observability(self, summary: dict[str, Any] | None) -> None:
        with self._lock:
            sanitized = _sanitize_value(summary or {}, max_len=120)
            self._tool_observability = sanitized if isinstance(sanitized, dict) else {}
        self._notify_listeners()

    def add_listener(self, listener: Callable[[], None]) -> None:
        with self._lock:
            if listener not in self._listeners:
                self._listeners.append(listener)

    def remove_listener(self, listener: Callable[[], None]) -> None:
        with self._lock:
            if listener in self._listeners:
                self._listeners.remove(listener)

    def snapshot(
        self,
        *,
        max_events: int = 8,
        max_log_lines: int = 120,
        allowed_sources: Iterable[str] | None = None,
        allowed_session_keys: Iterable[str] | None = None,
    ) -> dict[str, Any]:
        event_limit = max(0, int(max_events))
        log_limit = max(0, int(max_log_lines))
        source_filter = {
            _sanitize_text(source, max_len=80)
            for source in (allowed_sources or ())
            if str(source).strip()
        }
        session_filter = {
            _sanitize_text(session_key, max_len=120)
            for session_key in (allowed_session_keys or ())
            if str(session_key).strip()
        }
        with self._lock:
            recent_events = list(self._recent_events)
            if source_filter:
                recent_events = [
                    event
                    for event in recent_events
                    if str(event.get("source") or "") in source_filter
                ]
            if session_filter:
                recent_events = [
                    event
                    for event in recent_events
                    if str(event.get("session_key") or "") in session_filter
                ]
            return {
                "log_file_path": self._log_file_path,
                "recent_events": recent_events[:event_limit],
                "recent_log_lines": list(self._recent_log_lines)[-log_limit:] if log_limit else [],
                "tool_observability": dict(self._tool_observability),
                "event_count": len(recent_events),
                "log_line_count": len(self._recent_log_lines),
            }

    def _notify_listeners(self) -> None:
        with self._lock:
            listeners = list(self._listeners)
        for listener in listeners:
            try:
                listener()
            except Exception:
                continue


_STORE = RuntimeDiagnosticsStore()


def get_runtime_diagnostics_store() -> RuntimeDiagnosticsStore:
    return _STORE


def _fallback_log_target() -> Path:
    target = Path(tempfile.gettempdir()) / "bao" / "logs" / "desktop.log"
    target.parent.mkdir(parents=True, exist_ok=True)
    return target


def _resolve_log_target(log_file_path: str | Path | None) -> tuple[Path, str | None]:
    if log_file_path:
        target = Path(log_file_path).expanduser()
        try:
            target.parent.mkdir(parents=True, exist_ok=True)
            return target, None
        except OSError as exc:
            fallback = _fallback_log_target()
            return fallback, f"Configured log path unavailable: {target} ({exc})"

    from bao.config.loader import get_data_dir

    try:
        target = get_data_dir() / "logs" / "desktop.log"
        target.parent.mkdir(parents=True, exist_ok=True)
        return target, None
    except OSError as exc:
        fallback = _fallback_log_target()
        return fallback, f"Default data dir unavailable; using temp log dir ({exc})"


def _resolve_console_sink() -> Any | None:
    for candidate in (getattr(sys, "stderr", None), getattr(sys, "__stderr__", None)):
        if candidate is None:
            continue
        if callable(getattr(candidate, "write", None)):
            return candidate
    return None


def report_startup_failure(
    message: str,
    *,
    code: str = "desktop_startup_failed",
    details: dict[str, Any] | None = None,
) -> None:
    store = get_runtime_diagnostics_store()
    store.record_event(
        source="desktop_startup",
        stage="startup",
        message=message,
        level="error",
        code=code,
        retryable=False,
        details=details,
    )
    logger.error(message)


def configure_desktop_logging(log_file_path: str | Path | None = None) -> Path:
    target, fallback_reason = _resolve_log_target(log_file_path)
    store = get_runtime_diagnostics_store()
    store.set_log_file_path(target)

    logger.remove()
    logging.basicConfig(level=logging.WARNING, force=True)
    for name in ("httpcore", "httpx", "openai"):
        logging.getLogger(name).setLevel(logging.WARNING)

    console_format = "{time:HH:mm:ss} | {message}"
    file_format = "{time:YYYY-MM-DD HH:mm:ss.SSS} | {level:<8} | {message}"

    console_sink = _resolve_console_sink()
    if console_sink is not None:
        logger.add(console_sink, level="INFO", format=console_format)
    logger.add(
        str(target),
        level="DEBUG",
        rotation="5 MB",
        retention=5,
        encoding="utf-8",
        format=file_format,
    )
    logger.add(
        lambda message: store.append_log_line(str(message).rstrip("\n")),
        level="DEBUG",
        format=file_format,
    )
    if fallback_reason:
        store.record_event(
            source="desktop_startup",
            stage="logging",
            message=fallback_reason,
            level="warning",
            code="desktop_log_fallback",
            retryable=False,
            details={"log_file_path": str(target)},
        )
        logger.warning("{}", fallback_reason)
    return target



from __future__ import annotations

import json
from typing import Any, Iterable

from bao.agent.tools.base import Tool
from bao.runtime_diagnostics import RuntimeDiagnosticsStore, get_runtime_diagnostics_store


class RuntimeDiagnosticsTool(Tool):
    def __init__(
        self,
        store: RuntimeDiagnosticsStore | None = None,
        *,
        allowed_sources: Iterable[str] | None = None,
        pinned_session_key: str | None = None,
        allow_logs: bool = True,
        allow_tool_observability: bool = True,
    ) -> None:
        self._store = store or get_runtime_diagnostics_store()
        self._allowed_sources = tuple(
            str(source).strip() for source in (allowed_sources or ()) if str(source).strip()
        )
        self._pinned_session_key = str(pinned_session_key or "").strip()
        self._allow_logs = allow_logs
        self._allow_tool_observability = allow_tool_observability

    @property
    def name(self) -> str:
        return "runtime_diagnostics"

    @property
    def description(self) -> str:
        return "Read structured runtime diagnostics for recent internal failures and observability."

    @property
    def parameters(self) -> dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "max_events": {
                    "type": "integer",
                    "description": "Maximum number of recent diagnostic events to return.",
                    "minimum": 1,
                    "maximum": 20,
                },
                "include_logs": {
                    "type": "boolean",
                    "description": "Include a recent log tail excerpt when extra context is needed.",
                },
                "max_log_lines": {
                    "type": "integer",
                    "description": "Maximum number of recent log lines when include_logs is true.",
                    "minimum": 1,
                    "maximum": 200,
                },
                "source": {
                    "type": "string",
                    "description": "Optional source filter such as subagent, tool, provider, or agent_loop.",
                },
                "session_key": {
                    "type": "string",
                    "description": "Optional session/task filter when you need diagnostics for one scoped run.",
                },
            },
        }

    async def execute(self, **kwargs: Any) -> str:
        max_events = int(kwargs.get("max_events", 6) or 6)
        include_logs = bool(kwargs.get("include_logs", False))
        max_log_lines = int(kwargs.get("max_log_lines", 40) or 40)
        requested_source = str(kwargs.get("source") or "").strip()
        requested_session_key = str(kwargs.get("session_key") or "").strip()

        allowed_sources, blocked = self._resolve_sources(requested_source)
        if blocked:
            return self._render_payload(
                event_count=0,
                log_file_path=self._store.snapshot(max_events=0, max_log_lines=0)["log_file_path"],
                recent_events=[],
                tool_observability={},
                recent_log_lines=[],
                include_logs=include_logs,
                include_tool_observability=False,
            )

        effective_session_key = self._pinned_session_key or requested_session_key
        allowed_session_keys = [effective_session_key] if effective_session_key else []
        scoped_read = bool(allowed_sources or allowed_session_keys)

        snapshot = self._store.snapshot(
            max_events=max_events,
            max_log_lines=max_log_lines,
            allowed_sources=allowed_sources,
            allowed_session_keys=allowed_session_keys,
        )
        return self._render_payload(
            event_count=int(snapshot["event_count"]),
            log_file_path=str(snapshot["log_file_path"]),
            recent_events=list(snapshot["recent_events"]),
            tool_observability=dict(snapshot["tool_observability"]),
            recent_log_lines=list(snapshot["recent_log_lines"]),
            include_logs=include_logs and not scoped_read,
            include_tool_observability=not scoped_read,
        )

    def _resolve_sources(self, requested_source: str) -> tuple[list[str], bool]:
        if not self._allowed_sources:
            return ([requested_source] if requested_source else []), False
        if not requested_source:
            return list(self._allowed_sources), False
        if requested_source in self._allowed_sources:
            return [requested_source], False
        return [], True

    def _render_payload(
        self,
        *,
        event_count: int,
        log_file_path: str,
        recent_events: list[dict[str, Any]],
        tool_observability: dict[str, Any],
        recent_log_lines: list[str],
        include_logs: bool,
        include_tool_observability: bool,
    ) -> str:
        payload: dict[str, Any] = {
            "log_file_path": log_file_path,
            "event_count": event_count,
            "recent_events": recent_events,
        }
        if self._allow_tool_observability and include_tool_observability:
            payload["tool_observability"] = tool_observability
        if include_logs and self._allow_logs:
            payload["recent_log_lines"] = recent_log_lines
        return json.dumps(payload, ensure_ascii=False, indent=2)



from __future__ import annotations

_DEFAULT_API_BASE_BY_NAME: dict[str, str] = {
    "openai": "https://api.openai.com/v1",
    "openrouter": "https://openrouter.ai/api/v1",
    "deepseek": "https://api.deepseek.com/v1",
    "groq": "https://api.groq.com/openai/v1",
    "siliconflow": "https://api.siliconflow.cn/v1",
    "volcengine": "https://ark.cn-beijing.volces.com/api/v3",
    "dashscope": "https://dashscope.aliyuncs.com/compatible-mode/v1",
    "moonshot": "https://api.moonshot.ai/v1",
    "minimax": "https://api.minimax.io/v1",
    "zhipu": "https://open.bigmodel.cn/api/paas/v4",
    "vllm": "http://localhost:8000/v1",
    "ollama": "http://localhost:11434/v1",
    "lmstudio": "http://localhost:1234/v1",
    "aihubmix": "https://aihubmix.com/v1",
}


def get_default_api_base(provider_name: str) -> str:
    normalized = (provider_name or "").strip().lower().replace("-", "_")
    return _DEFAULT_API_BASE_BY_NAME.get(normalized, "")
